<?php
/**********************************************************************************************************************/
/*
Plugin Name: Electric Migration
Plugin URI:  https://www.electriccode.co.uk/wordpress-websites
Author:      Electric Code
Author URI:  https://www.electriccode.co.uk
Version:     1.1
Description: Makes migrating your Wordpress website from one web address to another as simple as possible. This plugin enables you to export your MySQL database while replacing all occurences of the old web address with the new one.
*/
/**********************************************************************************************************************/

	define( 'SQL_FILENAME', 'migrate.sql' );
	define( 'ROWS_PER_SEGMENT', 100 );

/**********************************************************************************************************************/

	$electric_migrate_current_url = home_url();
	$electric_migrate_new_url 	  = '';
	$electric_migrate_sql_path	  = '';
	$electric_migrate_result 	  = '';

/**********************************************************************************************************************/

	// Hook into init event
	function electric_migrate_init() {

		global $electric_migrate_current_url, $electric_migrate_new_url, $electric_migrate_result;

		// Hook into the admin menu event
		add_action( 'admin_menu', 'electric_migrate_admin_menu' );

		// If the migrate form has been posted back...
		if( isset( $_POST["electric_migrate_submit"] ) ){

			// Get the settings
			if( isset( $_POST["electric_migrate_current_url"] ) ){
				$electric_migrate_current_url = sanitize_text_field( $_POST["electric_migrate_current_url"] );
			};
			if( isset( $_POST["electric_migrate_new_url"] ) ){
				$electric_migrate_new_url = sanitize_text_field( $_POST["electric_migrate_new_url"] );
			};

			// Execute the migration
			$electric_migrate_result = electric_migrate_exec();
			if ( 'OK' == $electric_migrate_result ) {

				// Setup the redirection to set the download flag
		        $url = admin_url('tools.php?page=electric-migrate&download=true');
				header( 'Location:' . $url );
				die();

		    }

		}

		// If the download flag has been set send the sql file to the browser...
		if( isset( $_GET["download"] ) ) {
			$electric_migrate_result = 'OK';
			add_action( 'admin_init', 'electric_migrate_download_file');
		}

	}
	add_action( 'init' , 'electric_migrate_init' );

/**********************************************************************************************************************/

	// Provide CLI access
	if( class_exists( 'WP_CLI' ) ) {

		$to = function( $args, $assoc_args ) {

			global $electric_migrate_current_url, $electric_migrate_new_url;

			$electric_migrate_current_url = home_url();
			$electric_migrate_new_url = $args[0];
			$result = electric_migrate_exec();

			$path = plugin_dir_path( dirname( __FILE__ ) . '/electric-migrate.php' ) . SQL_FILENAME;
			$result .= ' ' . $path;
		    WP_CLI::success( $result );

		};
		WP_CLI::add_command( 'migrate', $to );

	}

/**********************************************************************************************************************/

	function electric_migrate_admin_menu() {

		// Add the management page to the tools admin menu
		add_management_page( 'Electric Migration', 'Electric Migration', 'manage_options', 'electric-migrate', 'electric_migrate_page' );

	}

/**********************************************************************************************************************/

	function electric_migrate_page() {

		global $electric_migrate_current_url, $electric_migrate_new_url, $electric_migrate_result;

		// If the current user lacks the required permissions to view options abort
		if ( !current_user_can( 'manage_options' ) )  {
			wp_die( __( 'You do not have sufficient permissions to access this page.' ) );
		}

		$html = '';

		// If a migration has been run output the result
		if( ''!=$electric_migrate_result ) {
			if( 'OK'!=$electric_migrate_result ) {
				$html.= '<div class="updated settings-error" id="setting-error-settings_updated">';
				$html.= '<p><strong>Migration failed:</strong></p><p>' . $electric_migrate_result . '</p></div>';
			} else {
				$html.= '<div class="updated settings-error" id="setting-error-settings_updated">';
				$html.= '<p><strong>Database migrated.</strong></p></div>';
			}
		}

		// Generate the plugin page
		$html .= <<<EOD

	<div class="wrap">
		<div id="icon-tools" class="icon32"></div>
		<h2>Electric Migration</h2><br/>
		<h4>Migrate your Wordpress site from one web address to another.</h4>
		<p>Enter the address of your new website and click 'Export Database' - the sql file for your new site should download automatically.</p>
		<form name="frmSettings" method="post">
			<table class="form-table">
				<tr>
					<th scope="row">Current address (URL)</th>
					<td><input type="text" id="electric_migrate_current_url" name="electric_migrate_current_url" size="40" value="$electric_migrate_current_url"/></td>
				<tr></tr>
					<th scope="row">New address (URL)</label></th>
					<td><input type="text" id="electric_migrate_new_url" name="electric_migrate_new_url" size="40" value=""/></td>
				</tr>
			</table>
			<br/>
			<input type="submit" value="Export Database" class="button button-primary" id="submit" name="electric_migrate_submit">
		</form>
	</div>
	<script type="text/javascript">
		jQuery(document).ready(function(){
			jQuery("#electric_migrate_new_url").focus();
		});
	</script>
EOD;

		// Output the plugin page
		echo $html;


	}

/**********************************************************************************************************************/

	function electric_migrate_exec() {

		global $electric_migrate_current_url, $electric_migrate_new_url, $electric_migrate_sql_path;

		// Initialise result
		$result = 'OK';

		// Get the WordPress globals relating to the database
		global $table_prefix, $wpdb;

		// Establish the path to this plugin
		$path = plugin_dir_path( dirname( __FILE__ ) . '/electric-migrate.php' );

		// Establish the path to the sql file
		$electric_migrate_sql_path = $path . SQL_FILENAME;

		// Before we go any further check that the plugin folder contains the .htaccess file
		// to prevent access to the sql files via the web for security reasons
		$access_file = $path.".htaccess";
		if( !file_exists( $access_file ) ) {
			file_put_contents( $access_file, "deny from all\n" );
		}

		// Create the sql file
		file_put_contents( $electric_migrate_sql_path, '' );

		// Start the sql file
		electric_migrate_write_to_sql_file( "# **********************************************************************" );
		electric_migrate_write_to_sql_file( "# WordPress Database Migration" );
		electric_migrate_write_to_sql_file( "# From " . $electric_migrate_current_url . " to " . $electric_migrate_new_url );
		electric_migrate_write_to_sql_file( "# Generated: " . date("l j. F Y H:i T") );
		electric_migrate_write_to_sql_file( "# Generated by Electric Migration" );
		electric_migrate_write_to_sql_file( "# Hostname: " . DB_HOST );
		electric_migrate_write_to_sql_file( "# Database: " . DB_NAME );
		electric_migrate_write_to_sql_file( "# **********************************************************************" );
		electric_migrate_write_to_sql_file( "#" );

		// Get all the tables in the database
        $tables = $wpdb->get_results( "SHOW FULL TABLES", ARRAY_N );

        // Process each table
		foreach ( $tables as $table ) {

			// Skip views
            if ( 'VIEW' == $table[1] ) continue;

            // Get table name
            $table = $table[0];

			// Increase script execution time-limit to 15 min for every table.
			if ( !ini_get('safe_mode' ) ) @set_time_limit(15*60);

			//Migrate the table to the sql file
			$result = electric_migrate_table_to_file( $table );
			if( 'OK'!=$result ) return $result;
		}

		return $result;
	}

/**********************************************************************************************************************/

	function electric_migrate_table_to_file( $table='' ) {

		global $wpdb;

		// Write the start of the sql
		electric_migrate_write_to_sql_file( "# **********************************************************************" );
		electric_migrate_write_to_sql_file( "# Table: " . electric_backquote( $table ) );
		electric_migrate_write_to_sql_file( "# **********************************************************************" );

		// Get the table structure
		$table_structure = $wpdb->get_results("DESCRIBE $table");
		if (! $table_structure ) {
			return "Error getting table details: " . electric_backquote( $table );
		}

		// Drop existing table
		electric_migrate_write_to_sql_file( "" );
		electric_migrate_write_to_sql_file( "# Delete any existing table " . electric_backquote( $table ) );
		electric_migrate_write_to_sql_file( "DROP TABLE IF EXISTS " . electric_backquote( $table ) . " ;");

		// Table structure
		electric_migrate_write_to_sql_file( "");
		electric_migrate_write_to_sql_file( "# Table structure of table " . electric_backquote( $table ) ) ;
		$create_table = $wpdb->get_results( "SHOW CREATE TABLE $table", ARRAY_N );
		if ( false === $create_table ) {
			electric_migrate_write_to_sql_file( "# Error with SHOW CREATE TABLE for " . $table );
			return 'Error with SHOW CREATE TABLE for ' . $table;
		}
		electric_migrate_write_to_sql_file( $create_table[0][1] . ' ;' );
		if ( false === $table_structure ) {
			electric_migrate_write_to_sql_file( "# Error getting table structure of  " . $table );
			return 'Error getting table structure of ' . $table;
		}

		// Table contents
		electric_migrate_write_to_sql_file( "" );
		electric_migrate_write_to_sql_file( '# Data contents of table ' . electric_backquote( $table ) );

		// Initialise arrays
		$defs = array();
		$ints = array();

		// Analyse table structure
		foreach ( $table_structure as $struct ) {

			if ( ( 0 === strpos( $struct->Type, 'tinyint' ) ) ||
				(0 === strpos(strtolower( $struct->Type ), 'smallint' ) ) ||
				(0 === strpos(strtolower( $struct->Type ), 'mediumint' ) ) ||
				(0 === strpos(strtolower( $struct->Type ), 'int' ) ) ||
				(0 === strpos(strtolower( $struct->Type ), 'bigint' ) )
				) {
					$defs[strtolower( $struct->Field )] = ( null === $struct->Default ) ? 'NULL' : $struct->Default;
					$ints[strtolower( $struct->Field )] = "1";
			}
		}

		// Batch
		$row_start = 0;
		$row_inc = ROWS_PER_SEGMENT;

		// Start processing the table data
		do {

			if ( !ini_get('safe_mode' ) ) @set_time_limit(15*60);

			// Get a segment of data
			$table_data = $wpdb->get_results( "SELECT * FROM $table LIMIT {$row_start}, {$row_inc}", ARRAY_A );

			// Prepare to process the segment of data
			$entries = 'INSERT INTO ' . electric_backquote( $table ) . ' VALUES (';
			$search = array("\x00", "\x0a", "\x0d", "\x1a");
			$replace = array('\0', '\n', '\r', '\Z' );

			// If we've got a segment of data
			if( $table_data ) {

				// Process each row of data
				foreach ( $table_data as $row ) {

					// Initialise the values
					$values = array();

					// Process each field
					foreach ( $row as $key => $value ) {

						if (isset( $ints[strtolower( $key )] ) && $ints[strtolower( $key )] ) {
							// make sure there are no blank spots in the insert syntax, yet try to avoid quotation marks around integers
							$value = ( null === $value || '' === $value ) ? $defs[strtolower( $key)] : $value;
							$values[] = ( '' === $value ) ? "''" : $value;
						} else {
							if (null === $value ) {
                                $values[] = 'NULL';
                            }
							else {
                                if ( is_serialized( $value ) && false !== ( $data = @unserialize( $value ) ) ) {
                                    if ( is_array( $data ) ) {
                                        array_walk_recursive( $data, 'electric_replace_array_values' );
                                    }
                                    elseif ( is_string( $data ) ) {
                                        $data = electric_apply_replaces( $data, true );
                                    }
                                    $value = serialize( $data );
                                } else {
                                    $value = electric_apply_replaces( $value );
                                }
                                $values[] = "'" . str_replace( $search, $replace, electric_sql_addslashes( $value ) ) . "'";
                            }
						}
					}

					electric_migrate_write_to_sql_file( $entries . implode(', ', $values) . ' ) ;' );

				}
				$row_start += $row_inc;
			}
		} while( ( count( $table_data ) > 0 ) );

		//Write the end of the sql
		electric_migrate_write_to_sql_file( "" );

		// Return success;
		return 'OK';

	}

/**********************************************************************************************************************/

	function electric_migrate_download_file() {

		// Clear execution time limit
        set_time_limit(0);

		// Establish the path to this plugin
		$path = plugin_dir_path( dirname( __FILE__ ) . '/electric-migrate.php' );

		// Establish the path to the sql file
		$electric_migrate_sql_path = $path . SQL_FILENAME;

		// If the file doesnt exist quit
		if( !file_exists( $electric_migrate_sql_path ) ) {
			wp_die( 'File "' . $electric_migrate_sql_path . '" not found.' );
		}

		// Generate a decent filename for the download
		$filename = home_url();
		$filename = str_replace('https://', '', $filename);
		$filename = str_replace('http://', '', $filename);
		$filename = str_replace('www.', '', $filename);
		$filename = str_replace('.', '-', $filename);

		// Send the file to the browser
        header( 'Content-Description: File Transfer' );
        header( 'Content-Type: application/octet-stream' );
        header( 'Content-Length: ' . filesize( $electric_migrate_sql_path ) );
        header( 'Content-Disposition: attachment; filename=migrate-' . $filename . '.sql' );
        readfile( $electric_migrate_sql_path );

        // We're all done
        exit;

	}

/**********************************************************************************************************************/

	function electric_migrate_write_to_sql_file( $sql = '' ) {

		global $electric_migrate_sql_path;

		$sql_line = $sql . "\n";
		file_put_contents( $electric_migrate_sql_path, $sql_line, FILE_APPEND);

	}

/**********************************************************************************************************************/

	function electric_backquote( $table_name ) {

		if (!empty( $table_name ) && $table_name != '*' ) {
			if (is_array( $table_name ) ) {
				$result = array();
				reset( $table_name );
				while(list( $key, $val) = each( $table_name ) )
					$result[$key] = '`' . $val . '`';
				return $result;
			} else {
				return '`' . $table_name . '`';
			}
		} else {
			return $table_name;
		}

	}

/**********************************************************************************************************************/

	function electric_replace_array_values( &$value, $key ) {

        if ( !is_string( $value ) ) return;
        $value = electric_apply_replaces( $value, true );

	}

/**********************************************************************************************************************/

    function electric_apply_replaces( $subject, $is_serialized = false ) {

		global $electric_migrate_current_url, $electric_migrate_new_url;

        $search = array( $electric_migrate_current_url );
        $replace = array( $electric_migrate_new_url );
        $new = str_replace( $search, $replace, $subject );

        return $new;

    }

/**********************************************************************************************************************/

	function electric_sql_addslashes( $a_string = '', $is_like = false ) {

		if( $is_like ) {
			$a_string = str_replace( '\\', '\\\\\\\\', $a_string );
		} else {
			$a_string = str_replace( '\\', '\\\\', $a_string );
		}

		return str_replace( '\'', '\\\'', $a_string );

	}

/**********************************************************************************************************************/
?>
